"""Module to fetch example data."""

from vargram import __version__

import os
import requests
import zipfile
import io

def get_specific_tag(username, reponame, tag):
    """Download a given release of the VARGRAM repo. 

    Parameters
    ----------
    username : str
        The owner of the VARGRAM repository.
    reponame : str
        The name of the VARGRAM repository.
    tag : str
        The specific release of VARGRAM.

    Returns
    -------
    requests.Response
        The Response object returned by a GET request to the URL of the release.
        
    """
    
    url = f'https://github.com/{username}/{reponame}/archive/refs/tags/{tag}.zip'
    response = requests.get(url)
    return response

def get_tag_names(username, reponame):
    """Gets all release tags of VARGRAM repository.

    Parameters
    ----------
    username : str
        The GitHub username or owner of the VARGRAM repository.
    reponame : str
        The name of the VARGRAM repository.

    Returns
    -------
    list
        List of all release tags.

    """

    url = f'https://api.github.com/repos/{username}/{reponame}/tags'
    response = requests.get(url)
    response.raise_for_status()
    return [tag['name'] for tag in response.json()]

def get(dir=''):
    """Downloads test data into target directory.
    
    Parameters
    ----------
    dir : str 
        Directory where test data will be downloaded.

    Returns
    -------
    None

    Raises
    ------
    requests.HTTPError
        If the GitHub request fails.
    """
    username = 'pgcbioinfo'
    reponame = 'vargram'

    # Define target directory where data will be downloaded
    if dir != '' and isinstance(dir, str): # Use user-provided directory 
        target_dir = dir 
        if not os.path.isdir(target_dir):
            os.makedirs(target_dir)
    else: # Or use current working directory by default
        target_dir = os.getcwd()

    # Get zipped data from VARGRAM repo based on installed package version
    downloaded_release = f"v{__version__}"
    all_tags = get_tag_names(username, reponame)
    if downloaded_release in all_tags: # i.e. the current version has not yet been pushed to the public repo
        all_tags.remove(downloaded_release)
    for tag in [downloaded_release] + all_tags:
        try:
            response = get_specific_tag(username, reponame, tag)
            response.raise_for_status()
            break
        except Exception:
            continue
        
    # Unzip repo
    with zipfile.ZipFile(io.BytesIO(response.content)) as zipped_repo:

        # Go through each archive member (i.e. filepaths) in zipfile
        # Skip members that aren't inside the test data directory (source_dir)
        source_dir = os.path.join(f"vargram-{tag[1:]}/","tests/test_data/")
        for path in zipped_repo.namelist():
            if path.startswith(source_dir) and  path != source_dir: 

                # Define target file path of member
                int_path = path[len(source_dir):] 
                target_path = os.path.join(target_dir, int_path)

                # Create member subdirectories
                if path.endswith('/'): 
                    if not os.path.exists(target_path):
                        os.mkdir(target_path)
                    continue
                
                # And extract files to target directory
                with zipped_repo.open(path) as source, open(target_path, "wb") as target:
                    target.write(source.read())
    
    if tag[1:] != __version__:
        print(f"Successfully downloaded vargram-{tag} (installed version: v{__version__}) test data to {target_dir}.")


"""Main module to generate the mutation profile."""

# Some term clarifications
# "group" -> gene, "stack" -> batch, "x" -> mutations

from . import _profile_renderer
from ..wranglers._nextclade_utils import parse_mutation, get_mutation_type
import matplotlib.pyplot as plt
import matplotlib.colors as mc
import numpy as np
import pandas as pd
from decimal import Decimal

def create_default_colors(num_color):
    """Creates default stack colors.

    Parameters
    ----------
    num_color : int
        The number of stacks to generate colors for.

    Returns
    -------
    list
        List of colors per stack.
    """
    # Default VARGRAM colors
    vggray = '#657C93'
    vgblue = '#009193'
    vgpink = '#E33E84'
    vgyellow = '#F09937'
    vgpurple = '#3A2B95'
    vgcolors = [vggray, vgblue, vgpink, vgyellow, vgpurple]

    # Use viridis cmap for large number of stacks/batches
    if num_color > 5:
        cmap_name = 'viridis'  
        cmap = plt.get_cmap(cmap_name)
        listed_cmap = cmap(np.linspace(0, 1, num_color))
        return [mc.to_hex(color) for color in listed_cmap]
    
    # Otherwise use VARGRAM colors
    match num_color:
        case 1:
            return [vgblue]
        case _:
            return vgcolors[:num_color]

def match_attributes(cds_name, cds_attribute_string):
    """Search attributes for a matching CDS name"""
    nextclade_cds_keys = ["Gene", "gene", "gene_name", "locus_tag",
                                "Name", "name", "Alias", "alias", 
                                "standard_name", "old-name", "product",
                                "gene_synonym", "gb-synonym", "acronym",
                                "gb-acronym", "protein_id", "ID"]
    attributes = cds_attribute_string.split(';')
    attributes_dict = {attribute.split('=')[0].strip(): attribute.split('=')[1].strip() for attribute in attributes}
    matched = False
    for key in nextclade_cds_keys:
        if key in attributes_dict.keys():
            if attributes_dict[key] == cds_name:
                matched = True

    return matched

class Profile():

    def __init__(self, wrangled_data):
        """Initializes Profile attributes."""
        self.data = wrangled_data["data"].copy() # User-provided data
        self.format = wrangled_data["format"] # Format of data (e.g. Nextclade, VARGRAM)
        self.annotation = wrangled_data.get("annotation") # Genome annotation file
        self.plotted_already = False # Flag for whether the actual figure has been created
        self.verbose = False # Flag for printing completion of a method call
        self.fig = plt.figure() # The profile Figure object
        self.closed = False # Flag for whether figure has been closed
        self.shown = False # Flag for whether the figure is shown
        self.key_called = False # Flag for whether key files have been provided
        self.struct = [] # The order or the structure of the groups/genes in the figure

    def process(self, **process_kwargs):
        """Creates data for plotting and structure.
        
        Returns
        -------
        pandas.DataFrame
            The processed DataFrame for plotting.

        """
        # Creating profile attributes
        for process_key in process_kwargs.keys():
            setattr(self, process_key, process_kwargs[process_key])
        
        # Assigning default group and stack titles
        if self.group_title == '':
           self.group_title = self.group
           if self.group == 'gene':
                self.group_title = self.group.capitalize()
        if self.stack_title == '':
           self.stack_title = self.stack
           if self.stack == 'batch':
                self.stack_title = self.stack.capitalize()

        # vargram output is user-input
        if self.format == 'vargram':
            self.data_for_plotting = self.data
            if self.verbose:
                print('** Processed data for plotting. **')
            return self.data_for_plotting

        # Checking if the default columns are present or if not, required arguments are provided
        required_data_attributes = ['x', 'group', 'stack']
        default_data_attributes = [self.x, self.group, self.stack]
        default_columns_detected = all(col in self.data.columns.tolist() for col in default_data_attributes)
        required_attributes_detected = all(arg in process_kwargs.keys() for arg in required_data_attributes) #if arg != 'threshold' and arg != 'ytype')
        if not default_columns_detected and not required_attributes_detected:
            raise ValueError(f"Expected columns or arguments not detected.")

        for process_key in process_kwargs.keys():
            setattr(self, process_key, process_kwargs[process_key])

        # Pivoting dataframe to get x counts
        # self.data -> data_pivoted
        data_pivoted = self.data.copy()
        if self.y == '': # Choosing what to base counts on
            values_for_counting = 'values_for_counting'
            data_pivoted[values_for_counting] = 1
        elif self.y != '':
            values_for_counting = self.y
        # Defining index columns
        index_columns = [self.group, self.x]
        data_pivoted = pd.pivot_table(data_pivoted, index=index_columns, columns=self.stack, values=values_for_counting, aggfunc="sum", fill_value=0).reset_index() 
        data_pivoted.rename_axis(None, axis=1, inplace = True)

        # Applying threshold, keeping only x
        # data_pivoted -> data_filtered
        self.stack_names = [stack for stack in data_pivoted.columns if stack not in index_columns]
        if len(self.stack_label) == 0: # Assigning stack_names as labels
            self.stack_label = self.stack_names
        data_filtered = data_pivoted.copy()
        for stack in self.stack_names:
            data_filtered[stack] = data_pivoted[stack].apply(lambda x: 0 if x < self.threshold else x)
        
        # Determining whether to normalize or not
        # weights vs. counts
        if self.ytype == '':
            if len(self.stack_names) > 1:
                self.ytype = 'weights'
            else:
                self.ytype = 'counts'
        if self.ylabel == '':
            self.ylabel = self.ytype.title()
        
        if self.ytype == 'weights':
            for stack in self.stack_names: # Using Decimal for precision
                data_filtered[stack] = data_filtered[stack].apply(lambda x: Decimal(str(x)))
                stack_sum = data_filtered[stack].sum()
                if stack_sum == 0:
                    data_filtered[stack] = data_filtered[stack] * Decimal('0')
                    continue
                data_filtered[stack] = data_filtered[stack] * Decimal('100') / Decimal(str(stack_sum))
                data_filtered[stack] = data_filtered[stack].apply(lambda x: x.quantize(Decimal('0.01'))) 
        
        # Summing x counts across all stacks
        data_filtered['sum'] = data_filtered[self.stack_names].sum(axis=1)
        if self.ytype == 'weights': # Converting Decimal objects back to float
            for col in self.stack_names + ['sum']:
                data_filtered[col] = data_filtered[col].apply(float)
        data_filtered = data_filtered[data_filtered['sum'] > 0]
        
        # Adding keys if provided
        # data_filtered -> self.data_for_plotting
        # data_filtered -> data_with_keys -> self.data_for_plotting
        if self.key_called:
            data_with_keys = pd.merge(data_filtered, self.key_data, on=[self.group, self.x], how='outer')
            data_with_keys.fillna(0, inplace=True)
            if data_filtered['sum'].dtype == np.int64:
                numerical_columns = list(data_with_keys.columns)[2:]
                for col in numerical_columns:
                    data_with_keys[col] = data_with_keys[col].astype(np.int64)
            data_with_keys.reset_index(drop=True, inplace=True)
            self.data_for_plotting = data_with_keys
        else:
            self.data_for_plotting = data_filtered

        # Sorting data based on mutation position
        if self.data_for_plotting.columns.size > 0 and self.data_for_plotting.shape[0] == 0:
            raise ValueError("Plot DataFrame has no rows. Lowering threshold might help.")
        formats_with_positions = ['nextclade_fasta', 'nextclade_delimited', '_test']
        if self.format in formats_with_positions:
            pos_index = self.data_for_plotting.columns.get_loc('mutation') + 1
            mutation_positions = self.data_for_plotting['mutation'].apply(lambda x: parse_mutation(x, 'position'))
            mutation_types = self.data_for_plotting['mutation'].apply(lambda x: get_mutation_type(x))
            self.data_for_plotting.insert(pos_index, 'position', mutation_positions)
            self.data_for_plotting.insert(pos_index+1, 'type', mutation_types)
            self.data_for_plotting['position'] = self.data_for_plotting['position'].astype(int)
            self.data_for_plotting.sort_values(by=[self.group, 'position'], inplace=True)
        else:
            self.data_for_plotting.sort_values(by=[self.group, self.x], inplace=True)
        self.data_for_plotting.reset_index(drop=True, inplace=True)
        
        # Getting data for calculating structure
        data_for_plotting = self.data_for_plotting.copy()
        self.group_names = data_for_plotting[self.group].unique().tolist()
        unique_counts = []
        for group in self.group_names:
            groups_df = data_for_plotting[data_for_plotting[self.group] == group]
            unique_counts.append(len(groups_df[self.x].unique()))

        # Checking obtained groups from struct
        if self.struct:
            unrecognized_groups = [col for row in self.struct for col in row  
                                   if col not in self.group_names ]
            if unrecognized_groups:
                raise ValueError(f"""Unrecognized genes/group names: {unrecognized_groups}.
                                 Make sure to remove unnecessary spaces or characters.""")
        
        # Getting data for struct
        self.data_for_struct = pd.DataFrame({self.group: self.group_names, 'count': unique_counts})
        self.data_for_struct.sort_values(by='count', ascending=False, inplace=True)
        self.data_for_struct.reset_index(drop=True, inplace=True)
        
        # Returning for vg.stat()
        if self.verbose:
            print('** Processed data for plotting. **')
        return self.data_for_plotting
        
    def key(self, **key_kwargs):
        """Obtain the keys."""
        self.key_called = True        
        self.key_data = key_kwargs['key_data']
        self.key_label = key_kwargs['key_labels']
        self.key_color = key_kwargs['key_colors']
        if self.verbose:
            print('** Processed key for profile. **')

    def aes(self, **aes_kwargs):
        """Set aesthetic attributes."""
        for aes_key in aes_kwargs.keys():
            setattr(self, aes_key, aes_kwargs[aes_key])

        # Turning single-string label and color to lists
        if isinstance(self.stack_label , str):
            self.stack_label = [self.stack_label]
        if isinstance(self.stack_color, str) and bool(self.stack_color):
            self.stack_color = [self.stack_color]
        
        if self.verbose:
            print('** Processed aesthetics for profile. **')
    
    def _get_gene_orders(self):
        """Obtain the orders of the genes based on start position from GFF file."""
        # Nextclade prioritizes the CDS over genes. See Nextclade documentation.
        # Obtaining CDS rows corresponding to CDS names from data
        data_cds_names = self.data_for_struct['gene'].tolist()
        gene_and_cds = self.annotation[(self.annotation['feature'] == 'gene') | (self.annotation['feature'] == 'CDS')]
        cds_attributes = gene_and_cds['attribute'].tolist()
        cds_groups = dict()
        for cds_name in data_cds_names:
            matched_rows = []
            for row, cds_attribute in enumerate(cds_attributes):
                if match_attributes(cds_name, cds_attribute):
                    matched_rows.append(row)
            cds_groups[cds_name] = matched_rows.copy()

        # Getting minimum start positions of each CDS group
        start_minimum = []
        for cds_name in data_cds_names:
            start_positions = []
            for row in cds_groups[cds_name]:
                start = gene_and_cds.iloc[row]['start']
                start_positions.append(start)
            start_minimum.append(int(min(start_positions)))

        # Ordering the CDS
        names_start = list(zip(data_cds_names, start_minimum))
        sorted_names_start = sorted(names_start, key=lambda x: x[1])
        ordered_cds_names =  [name for name,_  in sorted_names_start]  
        ordered_cds_start = [start for _,start in sorted_names_start] 
        self.ordered_genes = ordered_cds_names
    
    def struct_method(self, **struct_kwargs):
        """Obtain the structure for the profile (i.e. the groups to include)."""
        # Getting user input
        list_of_lists = isinstance(struct_kwargs['struct_key'], list) and all(isinstance(item, list) for item in struct_kwargs['struct_key'])
        if list_of_lists:
            self.struct = struct_kwargs['struct_key']
        elif isinstance(struct_kwargs['struct_key'], str):
            row_groups = struct_kwargs['struct_key'].split('/')
            self.struct = [col.split(',') for col in row_groups]
        else:
            raise ValueError("Unrecognized struct input. Expected list or string.")
        if self.verbose:
            print('** Processed struct. **')

    def plot(self):
        """Create the figure."""
        if self.verbose:
            print('** Plotting **')
        
        # Getting structure of the mutation profile grid
        if len(self.struct) != 0:
            pass # self.struct will be as specified by user 
        elif self.annotation is not None and self.order:
            self._get_gene_orders()
            self.struct = _profile_renderer.build_ordered_struct(self.data_for_struct,
                                                                 self.group,
                                                                 self.ordered_genes,
                                                                 self.flat)
        elif len(self.struct) == 0:
            self.struct = _profile_renderer.build_struct(self.data_for_struct, 
                                                         self.group, 
                                                         self.flat)  
                    
        # Creating profile grids
        grids_and_axes = _profile_renderer.build_profile_grid(self.struct, 
                                                              self.data_for_struct,
                                                              self.group,
                                                              self.key_called)
        label_grid =  grids_and_axes[0]
        legend_grid =  grids_and_axes[1]
        group_title_axes = grids_and_axes[2]
        barplot_axes = grids_and_axes[3]
        heatmap_axes = grids_and_axes[4]
        
        # Gathering aesthetic attributes
        if self.stack_color == '':
            self.stack_color = create_default_colors(len(self.stack_label))
        x_aes = [self.xticks_fontsize, self.xticks_rotation]
        y_aes = [self.yticks_fontsize, self.ylabel]
        group_aes = [self.group_title, self.group_fontsize]
        stack_aes = [self.stack_label, self.stack_color, self.stack_title]
        if self.key_called:
            key_aes = [self.key_fontsize, self.key_label, self.key_color]
        else:
            key_aes = [[], [], []]

        # Adjusting figure size or aspect ratio
        if self.figsize is not None:
            self.fig.set_size_inches(self.figsize[0], self.figsize[1], forward=True)
        
        # Building barplots
        group_labels = []
        _profile_renderer.build_profile(group_title_axes, 
                                        barplot_axes, 
                                        heatmap_axes,
                                        self.data_for_plotting,
                                        self.struct,
                                        self.group,
                                        self.x,
                                        self.fig,
                                        self.aspect,
                                        self.key_called,
                                        key_aes,
                                        self.stack_names,
                                        stack_aes,
                                        group_aes,
                                        group_labels,
                                        x_aes,
                                        y_aes)
        # Creating figure y-axis label
        _profile_renderer.build_yaxis_label(self.ylabel, 
                                            label_grid, 
                                            self.ylabel_fontsize)
        # Creating figure legend
        _profile_renderer.build_legend(legend_grid, 
                                       stack_aes, 
                                       group_aes, 
                                       group_labels,
                                       self.legtitle_fontsize,
                                       self.legentry_fontsize
                                       )
        
        if self.aspect is not None: 
            if self.aspect <= 0:
                raise ValueError(f"Aspect ratio must be positive but got {self.aspect}.")
            fig_width, fig_height = self.fig.get_size_inches()
            self.fig.set_size_inches(fig_width, (1/self.aspect)*fig_width, forward=True)
        
        # Tightening layout
        plt.tight_layout()

    def show(self):
        """Displays the generated figure."""
        if not self.plotted_already:
            self.plot()
            self.plotted_already = True
        if self.closed:
            show_fig = plt.figure()
            show_manager = show_fig.canvas.manager
            show_manager.canvas.figure = self.fig
            self.fig.set_canvas(show_manager.canvas)
        plt.show()
        self.shown = True
        if self.verbose:
            print('** Showed figure. **')

    def save(self, **save_kwargs):
        """Saves the generated figure or data."""
        if not self.plotted_already:
            self.plot()
            self.plotted_already = True
        file_extension = save_kwargs['fname'].lower().split('.')[-1]
        fig_extensions = ['png', 'pdf', 'jpg']
        # Saving CSV
        if file_extension not in fig_extensions:
            if 'fname' in save_kwargs:
                save_kwargs['path_or_buf'] = save_kwargs.pop('fname')
            if 'index' not in save_kwargs.keys():
                save_kwargs['index'] = False
            match file_extension:
                case 'csv':
                    save_kwargs['sep'] = ','
                case 'tsv':
                    save_kwargs['sep'] = '\t'
                case 'txt':
                    save_kwargs['sep'] = ' '
            self.data_for_plotting.to_csv(**save_kwargs)
            plt.close() # Closing to avoid figure from being shown when saving
            self.closed = True
            if self.verbose:
                print('** Saved data **')
        # Saving figure
        else:
            if self.shown:
                save_fig = plt.figure()
                save_manager = save_fig.canvas.manager
                save_manager.canvas.figure = self.fig
                self.fig.set_canvas(save_manager.canvas)
            if 'bbox_inches' not in save_kwargs.keys():
                save_kwargs['bbox_inches'] = 'tight'
            plt.savefig(**save_kwargs)
            plt.close() # Closing to avoid figure from being shown when saving
            self.closed = True
            if self.verbose:
                print('** Saved figure **')


"""Main module for reading user input data and processing with Nextclade"""

# Types of sequences input
# 1. A single FASTA containing multiple sequences (1 batch)
# 2. A directory containing N FASTA files, each containing multiple sequences (N batches)

from ._nextclade_utils import input_checker
from ._nextclade_cli import create_command, capture_output
import os
import pandas as pd
import tempfile
import shutil


def nextclade(**kwargs):
    """Takes input sequence FASTA files and transforms it 
    into a DataFrame to be plotted by VARGRAM.

    Parameters
    ----------
    seq : str
        FASTA file path of the sequences.
    ref : str
        FASTA file path of the reference sequence.
    gene : str
        GFF3 file path of the genome annotation.
    
    Returns
    -------
    None

    Raises
    ------
    ValueError
        If Nextclade analysis dataframe is empty.

    """
    input_checker(kwargs)
    try:
        # Creating secure temporary directory to store Nextclade analysis output file
        secure_analysis_dir = tempfile.mkdtemp(prefix="secure_analysis_dir")

        # Creating secure temporary directory to store Nextclade 
        # reference and genome annotation if Nextclade reference name is provided
        secure_ref_dir = tempfile.mkdtemp(prefix="secure_ref_dir")

        if os.path.isdir(kwargs["seq"]): # Case 1: A directory of FASTA files is provided
            files = os.listdir(kwargs["seq"])
            batches = [file for file in files if file.endswith(('.fasta', '.fa'))]
            if len(batches) == 0:
                raise ValueError("Directory contains no FASTA file. Ensure FASTA has extension '.fasta' or '.fa'.")

            # Getting Nextclade analysis output per FASTA batch and concatenating 
            # the dataframes
            seq_dir = kwargs["seq"]
            outputs = []
            kwargs_mod = kwargs
            for batch in batches:
                # Getting Nexctlade output
                kwargs_mod["seq"] = os.path.join(seq_dir,batch)
                nextclade_command, gene_path = create_command(input = kwargs_mod, secure_analysis_dir = secure_analysis_dir, secure_ref_dir = secure_ref_dir) 
                out = capture_output(nextclade_command)

                # Appending output
                batch_name = os.path.splitext(batch)[0]
                out.insert(0, 'batch', batch_name)
                outputs.append(out)
            nextclade_output = pd.concat(outputs, ignore_index=True)   
        else: # Case 2: One FASTA file provided
            nextclade_command, gene_path = create_command(input = kwargs, secure_analysis_dir = secure_analysis_dir, secure_ref_dir = secure_ref_dir) 
            nextclade_output = capture_output(nextclade_command)

            # Add batch name
            batch_name = os.path.basename(kwargs['seq'])
            batch_name = os.path.splitext(batch_name)[0]
            nextclade_output.insert(0, 'batch', batch_name)

        # Sorting by batch name and seq name:
        nextclade_output.sort_values(by=['batch', 'seqName'], inplace=True)
        nextclade_output.reset_index(drop=True, inplace=True)

        # Getting annotation
        gff_columns = ["seqname", "source", "feature", "start", "end", "score",
                       "strand", "frame", "attribute"]
        annotation = pd.read_csv(gene_path, sep="\t", comment="#", 
                                 header=None, names=gff_columns)

    # Remove created directories
    finally:
        if os.path.exists(secure_analysis_dir):
            shutil.rmtree(secure_analysis_dir)
        if os.path.exists(secure_ref_dir):
            shutil.rmtree(secure_ref_dir)

    # Removing sequences with warnings or errors
    nextclade_output.drop(nextclade_output.dropna(subset=['warnings']).index)
    nextclade_output.drop(nextclade_output.dropna(subset=['errors']).index)
    if nextclade_output.empty:
        raise ValueError("Nextclade analysis DataFrame is empty.")
    return nextclade_output, annotation


"""Main module for data wrangling."""

from ._nextclade import nextclade
from . import _nextclade_utils
import pandas as pd
import os

def read_table(table_object, nextclade_file=False):
    """Read tabular data (CSV, TSV or pandas.DataFrame)."""
    if isinstance(table_object, pd.DataFrame):
        table = table_object
    elif isinstance(table_object, str):
        ext = os.path.splitext(table_object)[1]
        if nextclade_file:
            csv_delim = ';'
        else:
            csv_delim = ','

        match ext:
            case '.csv':
                delimiter = csv_delim
            case '.tsv':
                delimiter = '\t'
            case '.gff':
                gff_columns = ["seqname", "source", "feature", "start", "end", "score",
                       "strand", "frame", "attribute"]
                table = pd.read_csv(table_object, sep="\t", comment="#", 
                                 header=None, names=gff_columns)
                return table
            case _:
                raise ValueError(f"Unrecognized file extension. Expecting .csv or .tsv file but got {ext}")
        
        table = pd.read_csv(table_object, delimiter=delimiter) 
    else:
        raise ValueError("Unrecognized object. Expecting pandas.DataFrame object or delimited text file (CSV, TSV, or GFF).")
    
    return table

class Wrangler():

    def __init__(self, wrangler_kwargs):
        """Determine which plot to process data for."""
        self.wrangler_kwargs = wrangler_kwargs
        self.plot = wrangler_kwargs['plot']
        self.format = wrangler_kwargs.get('format')
        self.wrangled_data = dict()
        del wrangler_kwargs['plot']
        if self.format is not None:
            del wrangler_kwargs['format']
        self.user_input = wrangler_kwargs

        match self.plot:
            case 'Profile':
                self._profile()
    
    def get_wrangled_data(self):
        """Combine and return wrangled data and other attributes."""
        if self.data.empty:
            raise ValueError("Wrangled data is empty.")
        
        self.wrangled_data["data"] = self.data
        self.wrangled_data["format"] =  self.format

        return self.wrangled_data
            
    def _profile(self):
        """Perform appropriate data wrangling method for Profile()."""
        profile_formats = ['nextclade_fasta', 'nextclade_delimited', 'vargram', 'delimited', '_test']

        # Assigning default format values
        if self.format is None:
            if 'seq' in self.user_input.keys():
                self.format = 'nextclade_fasta'
            elif 'data' in self.user_input.keys():
                self.format = 'nextclade_delimited'
        elif self.format not in profile_formats:
            raise ValueError(f"Unrecognized format: {self.format}.")

        # Wrangling data
        match self.format:
            case 'nextclade_fasta':
                nextclade_kwargs = {key: self.user_input[key] for key in ['seq', 'ref', 'gene'] if key in self.user_input.keys()}
                read_data, annotation = nextclade(**nextclade_kwargs)
                self.data = _nextclade_utils.process_nextclade(read_data)
                self.wrangled_data["annotation"] = annotation
            case 'nextclade_delimited':
                tabular_data = self.user_input['data']
                read_data = read_table(tabular_data, nextclade_file=True)
                if 'batch' not in read_data.columns:
                    read_data.insert(0, 'batch', 'my_batch')
                read_data.sort_values(by=['batch', 'seqName'], inplace=True)
                read_data.reset_index(drop=True, inplace=True)
                self.data = _nextclade_utils.process_nextclade(read_data)
            case _:
                tabular_data = self.user_input['data']
                read_data = read_table(tabular_data)
                self.data = read_table(tabular_data, nextclade_file=False)

        # Getting annotation data if provided
        annotation_provided = 'gene' in self.user_input.keys()
        annotation_not_yet_read = self.wrangled_data.get('annotation') is None
        if annotation_provided and annotation_not_yet_read:
            annotation = read_table(self.user_input['gene'])
            self.wrangled_data["annotation"] = annotation

        # Adding metadata if available
        if 'meta' in self.user_input.keys():
            # Getting columns to merge on 
            if 'join' not in self.user_input.keys():
                missing_join_error=("Missing 'join' argument. Provide column name(s) " 
                                    "to (outer) merge data and metadata on.")
                raise ValueError(missing_join_error)
            elif isinstance(self.user_input['join'], str): 
                join = [self.user_input['join']]
            else:
                join = self.user_input['join']
            metadata = self.user_input['meta']
            if len(join) != 1:
                metadata.rename(columns={join[1]:join[0]}, inplace=True)
                self.data = pd.merge(self.data, metadata, how='outer', on=join[0])
            elif self.format == 'nextclade_fasta' or self.format == 'nextclade_delimited': 
                # If 'seq' is provided, automatically join on nextclade_sequence_name
                nextclade_seqname = 'seqName'
                metadata.rename(columns={join[0]:nextclade_seqname}, inplace=True)
                self.data = pd.merge(self.data, metadata, 
                                      how='outer', on=nextclade_seqname)
            else: 
                self.data = pd.merge(self.data, metadata, how='outer', on=join)


"""Module for generating the unit plots, other plot elements like legends, 
and auxiliary functions."""

import matplotlib.colors as mc
import numpy as np
import pandas as pd


def check_xticks_overlap(xticks):
    """Checks whether there is an overlap between the x-axis tick labels.

    Parameters
    ----------
    xticks :matplotlib.axes.Axes.get_xticklabels
        x-axis tick labels.

    Returns
    -------
    bool
        True if there is an overlap, False otherwise.

    """
    bb1 = xticks[0].get_window_extent()
    bb2 = xticks[1].get_window_extent()
    return not (bb1.x1 < bb2.x0 or bb2.x1 < bb1.x0 or bb1.y1 < bb2.y0 or bb2.y1 < bb1.y0)


def build_group_barplot(ax_bar, categories, heights, 
                        floor, batch_color, suppress_spline, 
                        key_called, max_height, x_aes, y_aes):
    """Generates the individual barplot of a group.

    Parameters
    ----------
    ax_bar : matplotlib.axes.Axes
        The subplot where to place the barplot for a particular group.
    categories : pd.Series
        All x-axis values for a particular group.
    heights : pd.Series
        The corresponding y-axis values for a particular group.
    floor : list
        The bottom y-values at which to start plotting the bars.
    batch_color : str
        The color for this particular batch.
    suppress_spline : bool
        Determines whether the y-axis and left spine will be shown.
    max_height : int
        Maximum height of a stacked bar.
    x_aes : list
        Aesthetic attributes of the x-axis ticks and labels.
    y_aes : list
        Aesthetic attributes of the y-axis label and the y label itself.

    Returns
    -------
    None

    """
    # bar() settings
    width = 0.75
    edgecolor = 'black'
    linewidth = 1
    x_fontsize = x_aes[0]
    rotation = x_aes[1]
    y_fontsize = y_aes[0]

    # Creating barplot
    ax_bar.bar(x = categories,
           height = heights,
           bottom = floor,
           color = batch_color,
           width = width,
           edgecolor = edgecolor,
           linewidth = linewidth
           )
    
    # Removing spines
    ax_bar.spines["top"].set_visible(False)
    ax_bar.spines["right"].set_visible(False)
    ax_bar.spines["bottom"].set_visible(False)

    # Adjusting limits of x-axis and y-axis
    is_numeric_dtype = pd.api.types.is_numeric_dtype(categories)
    if not is_numeric_dtype:
        ax_bar.set_xlim(-0.5, len(categories) - 0.5)
    if max_height != 0: # Avoids UserWarning
        ax_bar.set_ylim(0.0, max_height) 
    else:
        ax_bar.set_ylim(0.0, 1.0)

    # Removing x-axis labels if key lineage is called
    if key_called:
        ax_bar.xaxis.set_visible(False)
    else:
        ax_bar.tick_params(axis='x', rotation=rotation, labelsize=x_fontsize)
        # Checking for x-axis overlap
        xticks = ax_bar.get_xticklabels()
        if len(xticks) > 1:
            margin = 0.1
            while check_xticks_overlap(xticks):
                # Get the current figsize
                fig_width, fig_height = ax_bar.figure.get_size_inches()
                # Update the figsize to accommodate the required margin
                new_fig_width = fig_width + margin
                ax_bar.figure.set_size_inches((new_fig_width,(3/4)*new_fig_width), forward=False)
    
    # Leaving y-axis and left spine depending on whether 
    # this corresponds to first group on the bar row
    if suppress_spline or max_height == 0:
        ax_bar.yaxis.set_visible(False)
        ax_bar.spines['left'].set_visible(False)
    else:
        ax_bar.spines['left'].set_linewidth(1.5)
        ax_bar.spines['left'].set_position(('outward', 5)) 
        ax_bar.yaxis.set_tick_params(width=1.5, labelsize=y_fontsize)
    return None


def create_colormap(cmap_name = "key", color = "#5E5E5E"):
    """Generate the colormap of a reference key.
    
    Parameters
    ----------
    cmap_name : str
        The name of the reference key.
    color : str
        Hex code of color to use for this key.
    
    Returns
    -------
    matplotlib.colors.LinearSegmentedColormap
        The colormap.

    """
    cmap_colors = ["#D5D5D5", color]
    return mc.LinearSegmentedColormap.from_list(cmap_name, cmap_colors)


def build_group_heatmap(ax_heat, group_xvalues, key_labels, key_fontsize, cmaps, suppress_label,x_aes):
    """Generates the individual heatmap of a reference key.

    Parameters
    ----------
    ax_heat : matplotlib.axes.Axes
        The subplot where to place the heatmap for a particular group.
    group_xvalues : pandas.DataFrame
        Summary counts for a particular group including key x values.
    key_labels : str
        The name of the reference key.
    key_fontsize : int
        The fontsize of the key labels.
    cmaps : matplotlib.colors.Colormap
        List of colormaps to use.
    suppress_label : bool
        Determines whether key label should be shown.
    x_aes : list
        Aesthetic attributes of the x-axis ticks and labels.

    Returns
    -------
    None

    """
    # Converting mutation names into binaries for imshow()
    xvalues_matrix = []
    for key_label in key_labels:
        xvalues_in_binary = group_xvalues[key_label]
        xvalues_matrix.append(xvalues_in_binary)

    # imshow() settings
    heatmap_border_color = 'black'
    heatmap_border_linewidth = 3
    heatmap_partition_color = 'white'
    heatmap_partition_linewidth = 1.5
    x_fontsize = x_aes[0]
    rotation = x_aes[1]

    # Creating heatmap
    reversed_cmaps=cmaps[::-1]
    mutation_names = group_xvalues['mutation']
    for i, row in enumerate(reversed(xvalues_matrix)):
        ax_heat.imshow([row], cmap=reversed_cmaps[i], vmin=0, vmax=1, extent=(-0.5, len(mutation_names)-0.5, i-0.5, i+0.5), aspect='auto')
    ax_heat.tick_params(axis='x', rotation=rotation, labelsize=x_fontsize)
    ax_heat.set_xticks(np.arange(len(mutation_names)))
    ax_heat.set_xticklabels(mutation_names)
    ax_heat.vlines(x=np.arange(0, len(mutation_names)-1)+0.5, ymin = -0.5, ymax = len(key_labels)-0.5, color = heatmap_partition_color, linewidth = heatmap_partition_linewidth)

    # Checking for x-axis overlap
    xticks = ax_heat.get_xticklabels()
    if len(xticks) > 1:
        margin = 0.1
        while check_xticks_overlap(xticks):
            # Get the current figsize
            fig_width, fig_height = ax_heat.figure.get_size_inches()
            # Update the figsize to accommodate the required margin
            new_fig_width = fig_width + margin
            ax_heat.figure.set_size_inches((new_fig_width,(3/4)*new_fig_width), forward=False)

    # Adding key lineage label
    if suppress_label:
        ax_heat.set_yticks([])
    else:
        ax_heat.set_yticks(list(range(len(key_labels))))
        ax_heat.set_yticklabels(key_labels[::-1])
        ax_heat.yaxis.set_tick_params(labelsize=key_fontsize)

    # Creating heatmap border
    ax_heat.vlines(x=-0.5, ymin=-0.5, ymax=len(key_labels)-0.5, color = heatmap_border_color, linewidth = heatmap_border_linewidth)
    ax_heat.vlines(x=len(mutation_names)-0.5, ymin=-0.5, ymax=len(key_labels)-0.5, color = heatmap_border_color, linewidth = heatmap_border_linewidth)
    ax_heat.hlines(y=np.linspace(-0.5, len(key_labels)-0.5, len(key_labels)+1), xmin= -0.5, xmax=len(mutation_names)-0.5, color = heatmap_border_color, linewidth = heatmap_border_linewidth)


def build_group_text(ax_text, group_name, fig_text, fontsize, aspect, group_labels):
    """Generates the group label above the barplot.

    Parameters
    ----------
    ax_text : matplotlib.axes.Axes 
        The subplot where to place the group name text for a particular group.
    group_name : str
        The text.
    fig_text : matplotlib.figure.Figure
        The Figure object of the entire VARGRAM bar plot.
    aspect : float
        The aspect ratio (width / height) of the Figure.
    fontsize : str or float
        The fontsize of the group text.
    group_labels : list
        List of group names that exceed the subplot box.

    Returns
    -------
    None

    """
    # text() settings
    fontsize = fontsize
    weight = 'bold'

    # Creating text
    xlims = ax_text.get_xlim()
    ylims = ax_text.get_ylim()
    t = ax_text.text(xlims[1]/2, 
                     ylims[1]/2, 
                     group_name, 
                     ha='center', 
                     va='center', 
                     transform=ax_text.transAxes, 
                     fontsize=fontsize, 
                     weight=weight)

    # Removing spines and ticks
    ax_text.set_yticks([])
    ax_text.set_xticks([])
    ax_text.spines['left'].set_linewidth(1.5)
    ax_text.spines["top"].set_linewidth(1.5)
    ax_text.spines["right"].set_linewidth(1.5)
    ax_text.spines["bottom"].set_linewidth(1.5)

    # Determining if text exceeds its subplot
    b = t.get_window_extent(renderer=ax_text.figure.canvas.get_renderer()).transformed(ax_text.transData.inverted())
    text_width = b.width
    text_height = b.height

    while text_height < 0.45: # Height is too small
        b = t.get_window_extent(renderer=ax_text.figure.canvas.get_renderer()).transformed(ax_text.transData.inverted())
        text_height = b.height
        fig_width, fig_height = ax_text.figure.get_size_inches()
        new_fig_height = 0.9*fig_height
        ax_text.figure.set_size_inches(fig_width, new_fig_height, forward=True)
    
    fig_width, fig_height = ax_text.figure.get_size_inches()
    while text_height > 0.8: # Height exceeds limit
        b = t.get_window_extent(renderer=ax_text.figure.canvas.get_renderer()).transformed(ax_text.transData.inverted())
        text_height = b.height
        fig_width, fig_height = ax_text.figure.get_size_inches()
        new_fig_height = 1.1*fig_height
        ax_text.figure.set_size_inches(fig_width, new_fig_height, forward=True)
    
    if text_width > 1: # Width exceeds limit
        margin = 1 # Increase only by small amount, otherwise use legend
        width_margin = margin
        # Get the current figsize
        fig_width, fig_height = ax_text.figure.get_size_inches()
        # Update the figsize to accommodate the required margin
        new_fig_width = fig_width + width_margin
        ax_text.figure.set_size_inches(new_fig_width, fig_height, forward=True)

    # Rechecking
    b = t.get_window_extent(renderer=fig_text.canvas.get_renderer()).transformed(ax_text.transData.inverted())
    text_width = b.width
    if text_width > 1.2:
        no_group_labels = len(group_labels)
        group_labels.append(group_name)
        t.set_text('{}'.format(no_group_labels+1))


def spine_remover(ax):
    """Removes the spine of an Axes object .
    
    Parameters
    ----------
    ax : matplotlib.Axes
        The Axes.

    Returns
    -------
    None

    """
    ax.set_yticks([])
    ax.set_xticks([])
    ax.xaxis.set_visible(False)
    ax.yaxis.set_visible(False)
    ax.spines["top"].set_visible(False)
    ax.spines["bottom"].set_visible(False)
    ax.spines["left"].set_visible(False)
    ax.spines["right"].set_visible(False)


"""Module that generates the bar structure, grid and plots its elements."""

from . import _profile_elements
import matplotlib.pyplot as plt
import matplotlib.gridspec as mg
import matplotlib.patches as mp
import matplotlib.text as mt
import copy


def build_ordered_struct(group_counts, group_attr, ordered_genes, 
                         flat=False, max_per_row=40):
    """Builds the structure based on the genes' start positions

    Parameters
    ----------
    group_counts : pandas.DataFrame
        The DataFrame containing groups and their unique no. of x data.
    group_attr : str
        The group data attribute.
    ordered_genes: list
        A list of ordered genes based on CDS start position.
    flat : bool, default : False
        Determines if groups should be plotted on only one row ("flat") or not ("compact").
    
    Returns
    -------
    list
        The structure of the plot where each row gives the list of groups for that row.
    """
    if flat is True:
        struct = [[gene for gene in ordered_genes]]
        return struct
    
    gg = group_counts[group_attr].tolist() # Gene names
    cc = group_counts['count'].tolist() # Gene counts
    ref_row_length = max(max_per_row, max(cc))
    struct = []
    row = []
    num_row = 0
    row_sum = 0
    for (i, gene) in enumerate(ordered_genes):
        gene_index = gg.index(gene)
        gene_count = cc[gene_index]

        if i == 0: # First gene is the first value in struct, no computation needed
            row.append(gene)
            row_sum += gene_count
            continue

        if row_sum < ref_row_length and num_row == 0:
            row.append(gene)
            row_sum += gene_count
        elif row_sum + gene_count <= ref_row_length:
            row.append(gene)
            row_sum += gene_count
        else:
            num_row += 1
            if row_sum > ref_row_length:
                ref_row_length = row_sum
            struct.append(row)
            row_sum = gene_count
            row = []
            row.append(gene)
    struct.append(row)

    return struct


def build_struct(group_counts, group_attr, flat=False, 
                 max_per_row=40):
    """Determines the optimum structure of the groups in the bar plot.

    Parameters
    ----------
    group_counts : pandas.DataFrame
        The DataFrame containing groups and their unique no. of x data.
    group_attr : str
        The group data attribute.
    flat : bool, default False
        Determines if groups should be plotted on only one row (if True).
    max_per_row : int, default 40
        Initial maximum number of x data per row.
    
    Returns
    -------
    list
        The structure of the plot where each row gives the list of groups for that row.

    """
    gg = group_counts[group_attr].tolist()
    cc = group_counts['count'].tolist()
    struct = [] # list of groups per row

    if flat:
        paired_counts = list(zip(gg, cc))
        descending_paired = sorted(paired_counts, key=lambda x: x[1], reverse=True)
        struct = [[group for group, _ in descending_paired]]
        return struct

    while len(gg) > 0: # Each iteration determines the groups for a row
        # Setting length of largest group as max_per_row
        # Since it is larger, it takes its own row
        largest_count = max(cc)
        if largest_count >= max_per_row: 
            max_per_row = largest_count
            largest_index = cc.index(largest_count)
            largest_group = gg[largest_index]
            struct.append([largest_group])
            cc.remove(largest_count)
            gg.remove(largest_group)
            continue

        # Of the remaining groups, find all that sum to less than or equal to max_per_row
        group_row = [gg[0]]
        current_sum = cc[0]
        for (i, group) in enumerate(gg):
            if i == 0:
                continue
            if current_sum + cc[i] <= max_per_row:
                group_row += [group]
                current_sum += cc[i]
        struct.append(group_row)

        # Remove these groups
        indices_to_remove = [i for i, group in enumerate(gg) if group in group_row]
        gg = [group for i, group in enumerate(gg) if i not in indices_to_remove]
        cc = [count for i, count in enumerate(cc) if i not in indices_to_remove]

    return struct


def build_profile_grid(struct, grid_width_counts, group_attr, 
                       key_called):
    """Creates the whole GridSpec objects on which to place the plots.

    Parameters
    ----------
    struct : list
        Contains the structure of the profile, i.e. group rows.
    grid_width_counts : pandas.Dataframe
        Contains x counts per group to be used for width ratios of grids
    group_attr : str
        The group attribute of the data
    key_called : bool
        Determines whether a key lineage was called or not.
    
    Returns
    -------
    mg.GridSpecFromSubplotSpec
        Grid for the figure y-axis label.
    mg.GridSpecFromSubplotSpec
        Grid for the stack (and group) legend.
    list
        List of Axes objects on which to plot the group title.
    list
        List of Axes objects on which to plot the bars.
    list
        List of Axes objects on which to plot the heatmap for key data.

    """
    # Main, outermost grid: 1 col for bar ylabel, 1 col for profile, 1 col for legend
    nrow = len(struct)
    bar_grid = mg.GridSpec(nrow, 3, width_ratios=[0.15, 21, 0.5])

    # Creating grid for the label and legend columns
    label_grid = mg.GridSpecFromSubplotSpec(1, 1, bar_grid[:, 0])
    legend_grid = mg.GridSpecFromSubplotSpec(2, 1, bar_grid[:, 2], hspace=0.1)

    # Getting maximum width
    width_max = 0
    all_width_ratios = []
    for (i, group_row) in enumerate(struct):
        group_row_counts = grid_width_counts[grid_width_counts[group_attr].isin(group_row)]
        ordered_group_row_counts = group_row_counts.set_index(group_attr).reindex(group_row).reset_index()
        width_ratios = ordered_group_row_counts['count'].tolist()
        all_width_ratios.append(width_ratios)
        if sum(width_ratios) > width_max:
            width_max = sum(width_ratios)

    # Creating grids for each row of the profile
    modified_group_rows = copy.deepcopy(struct)
    group_row_grids = []
    for (i, group_row) in enumerate(modified_group_rows):
        # Determining if row is not full
        # Then, add a filler group
        row_length = len(group_row)
        if sum(all_width_ratios[i]) < width_max:
            row_length += 1
            filler_length = width_max-sum(all_width_ratios[i])
            all_width_ratios[i].append(filler_length)
            modified_group_rows[i].append('filler')
        
        # Creating row grids
        if key_called:
            height_ratios = [1,7,1.5]
            group_row_grid = mg.GridSpecFromSubplotSpec(3, row_length, bar_grid[i, 1], 
                                                       width_ratios=all_width_ratios[i], 
                                                       height_ratios=height_ratios,
                                                       wspace=0.1,
                                                       hspace=0.1)
        else:
            height_ratios = [1,8.5]
            group_row_grid = mg.GridSpecFromSubplotSpec(2, row_length, bar_grid[i, 1], 
                                                       width_ratios=all_width_ratios[i],
                                                       height_ratios=height_ratios, 
                                                       wspace=0.1,
                                                       hspace=0.1)
        group_row_grids.append(group_row_grid)

    # Creating the axes
    group_title_axes = []
    group_x_axes = []
    group_key_axes = []
    for (i, group_row) in enumerate(modified_group_rows):
        group_row_grid = group_row_grids[i]
        for (j, group) in enumerate(group_row):
            if group == 'filler':
                continue
            group_title_grid = group_row_grid[0, j]
            group_x_grid = group_row_grid[1, j]
            
            # Creating subplot for group titles
            group_title_ax = plt.subplot(group_title_grid)

            # Creating subplot for the group barplots
            if j == 0:
                group_x_ax = plt.subplot(group_x_grid)
            else:
                first_group_index = len(group_x_axes) - j
                group_x_ax = plt.subplot(group_x_grid, sharey=group_x_axes[first_group_index])

            group_title_axes.append(group_title_ax)
            group_x_axes.append(group_x_ax)

            # Creating subplot for the group key x data
            if key_called:
                group_key_grid = group_row_grid[2, j]
                group_key_ax = plt.subplot(group_key_grid)
                group_key_axes.append(group_key_ax)      
    
    return label_grid, legend_grid, group_title_axes, group_x_axes, group_key_axes


def build_profile(group_title_axes, group_bar_axes, group_key_axes, 
                  barplot_data, struct, group_attr, 
                  x_attr, fig, aspect, 
                  key_called, key_aes, stack_names, 
                  stack_aes, group_aes, group_labels, x_aes, 
                  y_aes):
    """Generates the full profile including labels on the defined grids.

    Parameters
    ----------
    group_title_axes : list
        List of Axes objects on which to plot the group title.
    group_bar_axes : list
        List of Axes objects on which to plot the bars.
    group_key_axes : list
        List of Axes objects on which to plot the heatmap for key data.
    barplot_data : pandas.Dataframe
        The DataFrame containing summary x counts per stack including key values.
    struct : list
        The structure of the barplot/mutation profile.
    group_attr : str
        The column of the groups.
    x_attr : str
        The column of the x values.
    fig : matplotlib.figure.Figure
        The Figure object of the entire VARGRAM bar plot.
    aspect : float
        The aspect ratio (width / height) of the Figure.
    key_called : bool
        Determines whether a key was called or not.
    key_aes : list
        Aesthetic attributes of the key heatmap.
    stack_names : list
        The names of the stacks (from the data provided).
    stack_aes : list
        Aesthetic attributes of the stacks including 
        stack labels (may be user-provided) to put on the legend.
    group_aes : list
        Aesthetic attributes of the groups.
    group_labels : list
        List of group labels that exceed the subplot box.
    x_aes : list
        Aesthetic attributes of the x-axis ticks and labels.
    y_aes : list
        Aesthetic attributes of the y-axis label and the y label itself.
    
    Returns
    -------
    None

    """
    # Defining aesthetic attributes
    stacks = stack_names
    stack_colors = stack_aes[1]
    group_fontsize = group_aes[1]
    key_fontsize = key_aes[0]
    key_labels = key_aes[1]
    key_colors = key_aes[2]

    # Generating colormaps for each key lineage
    heat_cmaps = [] 
    for key_color in key_colors:
        heat_cmap = _profile_elements.create_colormap(color=key_color)
        heat_cmaps.append(heat_cmap)

    # Getting maximum stacked bar height across each group
    max_bar_heights = []
    for group_row in struct:
        max_bar_height = 0
        for group in group_row:
            group_barplot_data = barplot_data[barplot_data[group_attr] == group]
            if max(group_barplot_data['sum']) > max_bar_height:
                max_bar_height = max(group_barplot_data['sum'])
        group_row_max_heights = [max_bar_height]*len(group_row)
        max_bar_heights += group_row_max_heights

    # Flattening structure and getting first group on each row
    flattened_groups = [item for group_row in struct for item in group_row]
    first_groups = [group_row[0] for group_row in struct]

    # Going over each group and building the barplot, heatmap and group title
    for (i, group) in enumerate(flattened_groups):
        group_barplot_data = barplot_data[barplot_data[group_attr] == group]

        # Determining whether to suppress splines
        if group in first_groups:
            suppress_spline = False
            suppress_label = False
        else:
            suppress_spline = True
            suppress_label = True

        # Adding key x data for group
        if key_called:
            ax_heat = group_key_axes[i]
            heat_cmap = _profile_elements.create_colormap()
            _profile_elements.build_group_heatmap(ax_heat, 
                                                  group_barplot_data, 
                                                  key_labels, 
                                                  key_fontsize, 
                                                  heat_cmaps, 
                                                  suppress_label,
                                                  x_aes)
        
        # Adding text for group
        ax_text = group_title_axes[i]
        _profile_elements.build_group_text(ax_text, 
                                           group, 
                                           fig, 
                                           group_fontsize, 
                                           aspect,
                                           group_labels)

        # Creating unit barplot for group
        ax_bar = group_bar_axes[i]
        floor = [0]*len(group_barplot_data)
        for (batch, color) in zip(stacks, stack_colors):
            _profile_elements.build_group_barplot(ax_bar, 
                                                  group_barplot_data[x_attr], 
                                                  group_barplot_data[batch], 
                                                  floor,
                                                  color, 
                                                  suppress_spline, 
                                                  key_called, 
                                                  max_bar_heights[i],
                                                  x_aes, 
                                                  y_aes)
            floor += group_barplot_data[batch]            


def build_yaxis_label(label, label_grid, label_fontsize):
    """Generates the label of the figure y-axis.
    
    Parameters
    ----------
    label : str
        The y-axis label.
    label_grid : mg.GridSpecFromSubplotSpec
        Grid for the figure y-axis label.
    label_fontsize : str or float
        The font size of the y-axis label.
    
    Returns
    -------
    None

    """
    # text() settings
    ax_label = plt.subplot(label_grid[:, 0])

    # Creating label
    xlims = ax_label.get_xlim()
    ylims = ax_label.get_ylim()
    ax_label.text(xlims[1]/2, 
                  ylims[1]/2, 
                  label, 
                  ha='center', 
                  va='center', 
                  transform=ax_label.transAxes, 
                  fontsize=label_fontsize,
                  rotation=90)

    # Removing spines and ticks
    ax_label.set_yticks([])
    ax_label.set_xticks([])
    ax_label.spines["top"].set_visible(False)
    ax_label.spines["bottom"].set_visible(False)
    ax_label.spines["left"].set_visible(False)
    ax_label.spines["right"].set_visible(False)


def build_legend(legend_grid, stack_aes, group_aes, 
                 group_labels, title_fontsize, entry_fontsize):
    """Generates the label of the figure y-axis.
    
    Parameters
    ----------
    legend_grid : mg.GridSpecFromSubplotSpec
        Grid for the stack (and group) legend.
    stack_aes : list
        Aesthetic attributes of the stack legend.
    group_aes : str
        Aesthetic attributes of the group legend.
    group_labels : list
        List of group labels.
    title_fontsize : str or float
        Font size of the legend titles.
    entry_fontsize : str or float
        Font size of the legend entries.

    Returns
    -------
    None

    """
    # legend() settings
    if len(group_labels) == 0:
        ax_batch_legend = plt.subplot(legend_grid[:, 0])
    else:
        ax_batch_legend = plt.subplot(legend_grid[0, 0])
        ax_group_legend = plt.subplot(legend_grid[1, 0])
    stack_label = stack_aes[0]
    stack_color = stack_aes[1]
    stack_title = stack_aes[2]
    group_title = group_aes[0]
    frameon=False
    alignment='left'
    if len(group_labels) == 0:
        stack_loc='center'
        bbox_anchor = (0.5, 0.5)
    else:
        stack_loc='lower left'
        bbox_anchor = (-0.25,0)#(-0.5, 0)
    group_loc='upper left'

    # Setting batch legend handles
    batch_legend_handles = [mp.Patch(color=color, label=label) for color, label in zip(stack_color, stack_label)]

    # Creating batch legend
    ax_batch_legend.legend(handles=batch_legend_handles, 
                           title=stack_title, 
                           title_fontsize=title_fontsize,
                           fontsize=entry_fontsize, 
                           frameon=frameon, 
                           alignment=alignment, 
                           loc=stack_loc, 
                           bbox_to_anchor=bbox_anchor, 
                           borderaxespad=0)

    # Removing batch ax spines and ticks
    _profile_elements.spine_remover(ax_batch_legend)

    if len(group_labels) != 0:
        # Creating handles
        legend_handles = [Text(str(i+1)) for i in range(len(group_labels))]

        # Creating group legend
        ax_group_legend.legend(legend_handles, 
                               group_labels, 
                               title=group_title,
                               title_fontsize=title_fontsize, 
                               fontsize=entry_fontsize, 
                               frameon=frameon, 
                               alignment=alignment, 
                               loc=group_loc, 
                               bbox_to_anchor=(-0.25,1),#(-0.5, 1), 
                               borderaxespad=0,
                               handler_map={handle: TextHandler() for handle in legend_handles},
                               handletextpad=0.5,
                               labelspacing=0.9)
        
        # Removing group ax spines and ticks
        _profile_elements.spine_remover(ax_group_legend)


class Text(object):

    def __init__(self, text):
        self.text = text


class TextHandler(object):
    """Creates the text patch for the group legend."""

    def legend_artist(self, legend, text_handle, fontsize, handlebox):
        x0, y0 = handlebox.xdescent, handlebox.ydescent
        width, height = handlebox.width, handlebox.height
        patch = mt.Text(x=width/4, 
                        y=0, 
                        text=text_handle.text, 
                        bbox=dict(facecolor='none', boxstyle='Square'),
                        weight='bold',
                        verticalalignment=u'baseline', 
                        horizontalalignment=u'left', multialignment=None, 
                        fontproperties=None, linespacing=None, 
                        rotation_mode=None)
        handlebox.add_artist(patch)

        return patch


"""Tests whether profile plot is correct."""

from vargram.plots._profile_renderer import build_struct, build_profile_grid
import random
import pandas as pd


class TestProfileRenderer:

    def setup_method(self):
        # Creating group_counts dataframe
        counts = [50,48,29,10,1,20,2] # Counts are counts of unique mutations per gene
        self.num_genes = len(counts)
        genes = []
        for i in range(self.num_genes):
            genes.append(f'GENE_{i+1}')

        self.gene_counts = pd.DataFrame({
            'my_genes':genes,
            'count':counts
        })

        self.struct = [['GENE_1'], ['GENE_2', 'GENE_5'], ['GENE_3','GENE_4','GENE_7'], ['GENE_6']]
        self.axes = build_profile_grid(self.struct, self.gene_counts, 'my_genes', 'False')

        data_max = max(counts)
        self.predefined_max = 40
        self.max = max(data_max, self.predefined_max)

    def test_struct(self):
        """Asserts that struct algorithm is followed.
        
        Algorithm: Per row, take largest. If len(largest) > max, max = len(largest).
        If the next number added to current sum > max, exclude. Otherwise, include in row.
        Order matters. (In source, the count column is sorted first.)
        """
        expected = self.struct
        result = build_struct(self.gene_counts, 'my_genes', max_per_row=self.predefined_max)

        assert result == expected

    def test_num_axes(self):
        """Number of axes must be equal to number of genes."""

        num_group_title_axes = len(self.axes[2])
        num_barplot_axes = len(self.axes[3])
        num_heatmap_axes = len(self.axes[4])

        num_axes = [num_group_title_axes, num_barplot_axes, num_heatmap_axes]
        equal_to_num_genes = num_axes.count(self.num_genes) == len(num_axes)

        assert equal_to_num_genes == True

    def test_bar_width_ratios(self):
        """The width ratios of each bar ax must sum to the maximum row length (in terms of number of mutations)."""
        sum_width_ratios = []
        for bar_ax in self.axes[3]:
            bar_grid = bar_ax.get_gridspec()
            width_ratios = bar_grid.get_width_ratios()
            sum_width_ratios.append(sum(width_ratios))

        equal_to_max = sum_width_ratios.count(self.max) == len(sum_width_ratios)

        assert equal_to_max == True


"""Module to perform auxiliary tasks for nextclade function."""

import re
import subprocess
import os


def check_file_extension(file_path, valid_extensions):
    """Checks for the validity of the file extension.
    
    Parameters
    ----------
    file_path : str
        File path to be checked.
    valid_extensions : list
        List of valid extensions to be checked against.

    Returns
    -------
    file_extension : str 
        Extension of the file path provided.
    validity : bool
        Truth value of validity of the file extension.

    """
    file_extension = file_path.lower().split('.')[-1]
    if file_extension in valid_extensions:
        return file_extension, True
    else:
        return file_extension, False


def check_reference(ref):
    """Checks whether provided reference name is in Nextclade.
    
    Parameters
    ----------
    ref : str 
        The dataset name.

    Returns
    -------
    None

    Raises
    ------
    ValueError
        If Nextclade reference dataset name is not recognized.

    """
    dataset_names_only = subprocess.check_output(['nextclade', 'dataset', 'list', '--only-names']).decode(encoding='utf-8').split()  
    dataset_list_full = subprocess.check_output(['nextclade', 'dataset', 'list']).decode(encoding='utf-8')
    shortcut_parentheses = r'\(shortcuts:(.*?)\)' # getting all shortcuts enclosed in parentheses
    shortcut_quotes = r'"(.*?)"' # getting the individual shortcuts enclosed in double quotes
    dataset_shortcuts = re.findall(shortcut_parentheses, dataset_list_full)
    dataset_shortcuts = ' '.join(dataset_shortcuts)
    dataset_shortcuts = re.findall(shortcut_quotes, dataset_shortcuts)

    if ref not in dataset_names_only and ref not in dataset_shortcuts:
        try:
            name_not_recognized = ("Nextclade reference name not recognized. " 
                                    "See valid names and shortcuts below.")
            raise ValueError(name_not_recognized)
        except ValueError as e:
            dataset_list_command = ['nextclade', 'dataset', 'list']
            subprocess.run(dataset_list_command, check=True)
    return None


def input_checker(kwargs):
    """Takes the keyword arguments of nextclade() and checks for errors.

    Parameters
    ----------
    kwarg : dict 
        Set of keyword arguments passed to nextclade().

    Returns
    -------
    None

    Raises
    ------
    ValueError
        If 'seq' and 'ref' are not provided together.
        If 'gene' is not provided when 'ref' is not a Nextclade reference dataset name.
        If file extension is not expected.
    TypeError
        If an unexpected keyword is provided.

    """
    # Raise error for unknown keys
    expected_keys = {"seq", "ref", "gene"}#, "meta"}
    unexpected_keys = set(kwargs.keys()) - expected_keys
    if unexpected_keys:
        raise TypeError(f"Unexpected keyword(s): {', '.join(unexpected_keys)}")
    
    # Raise error when "seq" is provided but "ref" isn't 
    if "seq" in kwargs and "ref" not in kwargs:
        raise ValueError("Path of reference sequence is not provided.")
    
    # Raise error when "ref" is provided but "seq" isn't
    if "ref" in kwargs and "seq" not in kwargs:
        raise ValueError("Path of sequences is not provided.")  
    
    # Raise error when "gene" is not provided and "ref" is a file
    if "gene" not in kwargs and os.path.isfile(kwargs["ref"]):
        raise ValueError("Genome annotation is not provided.")   
    
    # Checking extensions
    for key in kwargs.keys():
        valid_extensions = {"seq": ["fa", "fasta"], 
                                "ref": ["fa", "fasta"],
                                "gene": ["gff", "gff3"],
                                "meta": ["csv", "tsv"]}
        file_type = {"seq": "sequence",
                         "ref": "reference",
                         "gene": "gene annotation",
                         "meta": "metadata"}
        if key == "seq" and not os.path.isfile(kwargs["seq"]):
            continue
        if key == "ref" and not os.path.isfile(kwargs["ref"]):
            check_reference(kwargs["ref"]) # Check if provided reference name is recognized instead
            continue
        file_extension, validity = check_file_extension(kwargs[key], valid_extensions[key])
        if not validity:
            raise ValueError(f"Unsupported {file_type[key]} file format: {file_extension}")


def parse_mutation(aa_mutation, part):
    """Gets the gene or position from the mutation. Can also remove the gene prefix of the mutation.

    Parameters
    ----------
    aa_mutation : str
        The amino acid mutation as given by the nomenclature of Nextclade,
        e.g. ORF1b:G662S.
    part : str
        The part (gene name or position) of the mutation to be retrieved. 
        Part can also indicate removal.

    Returns
    -------
    str 
        The gene name or position on which the mutation occurs. 
        Or the mutation without the gene prefix.
    
    Raises
    ------
    ValueError
        If position or gene is not parsed.

    """
    # Removing prefix
    if part == 'gene_removal':
        pattern = r'^[^:]+:'
        stripped = re.sub(pattern, '', aa_mutation)
        return stripped
    
    # Patterns for the gene name and position
    if part == 'gene':
        pattern = r'^([^:]+):'
    if part == 'position':
        pattern = r'(\d+)'

    # Obtaining the match
    match = re.findall(pattern, aa_mutation)
    if not match:
        raise ValueError(f"Failed to parse mutation for {part}: '{aa_mutation}'.")
    retrieved  = match[-1]
    return retrieved


def get_mutation_type(mutation):
    """Creates a new column for mutation type.
    
    Parameters
    ----------
    mutation : str
        The gene-stripped mutation name.

    Returns
    -------
    str
        The type of mutation.

    """
    if '-' in mutation:
        return 'del'
    elif ':' in mutation:
        return 'in'
    else:
        return 'sub'


def process_nextclade(nextclade_output):
    """Gets the unique mutations and their individual counts.

    Parameters
    ----------
    nextclade_output : pandas.DataFrame
        A subset (for multiple batches) of the DataFrame produced by nextclade().
    
    Returns
    -------
    pandas.DataFrame
        A DataFrame of unique mutations and their counts.
    
    Raises
    ------
    ValueError
        If created mutation column of analysis DataFrame is empty.

    """
    # Nextclade columns
    aa_sub = 'aaSubstitutions'
    aa_del = 'aaDeletions'
    aa_ins = 'aaInsertions'

    # Compiling all amino acid mutations (substitution, deletion, insertion) into a single column
    single_column = nextclade_output.copy()
    
    # Turn all NA mutations into empty strings
    single_column[aa_sub] = single_column[aa_sub].fillna('')
    single_column[aa_del] = single_column[aa_del].fillna('')
    single_column[aa_ins] = single_column[aa_ins].fillna('')

    # Per row, join the mutations into one (comma-separated) string in one column
    single_column['mutation'] = single_column.apply(lambda row: ','.join(filter(None, [row[aa_sub], row[aa_del], row[aa_ins]])), axis=1)
    
    # Remove rows with empty strings (i.e. sequence has no mutation at all)
    single_column = single_column[single_column['mutation'] != '']
    if single_column.empty:
        raise ValueError("Processed Nextclade analysis DataFrame is empty. Potentially sequences do not have any mutation.")
    
    # Generating a row per mutation
    # Each row therefore may not be unique
    single_column['mutation'] = single_column['mutation'].apply(lambda x: x.split(',')).copy()
    exploded = single_column.explode('mutation')

    # Creating gene column and removing gene prefix of mutations
    exploded['gene'] = exploded['mutation'].apply(lambda x: parse_mutation(x, 'gene'))
    exploded['mutation'] = exploded['mutation'].apply(lambda x: parse_mutation(x, 'gene_removal'))

    # Rearranging
    exploded.reset_index(drop=True, inplace=True)
    processed_nextclade = exploded[['batch', 'gene', 'mutation']]
    return processed_nextclade


"""
Tests whether VARGRAM output data is the same when equivalent inputs are given.
Nextclade CLI must be installed.
"""

from vargram import vargram

class TestFileInput():

    def setup_method(self):
        """Obtain output data for equivalent inputs."""

        # Using local data
        vg = vargram(seq='tests/test_data/sequences', 
                     ref='tests/test_data/sc2_wuhan_2019.fasta', 
                     gene='tests/test_data/sc2.gff')
        vg.profile(threshold=1, ytype='counts')
        self.local_ref = vg.stat()

        # Using VARGRAM output
        vg = vargram(data='tests/test_data/summary/omicron_summary_data_t1.csv',
                     format='vargram')
        vg.profile()
        self.stat = vg.stat()

        # Using Nextclade reference
        vg = vargram(seq='tests/test_data/sequences', 
                     ref='sars-cov-2', ytype='counts')
        vg.profile(threshold=1, ytype='counts')
        self.online_ref = vg.stat()

        # Using Nextclade CLI analysis output
        vg = vargram(data='tests/test_data/analysis/omicron_analysis_cli.tsv')
        vg.profile(threshold=1, ytype='counts')
        self.analysis_file = vg.stat()

    def test_stat(self):
        """Output using local data must equal to output when VARGRAM analysis output is provided as input."""
        assert self.local_ref.equals(self.stat)

    def test_ref(self):
        """Output using local reference must equal output using downloaded reference."""
        assert self.local_ref.equals(self.online_ref)
    
    def test_analysis(self):
        """Output using local data must equal output using Nextclade CLI analysis output as input."""
        shared_columns = ['gene', 'mutation', 'position', 'type', 'sum']
        sh_local_ref = self.local_ref[shared_columns]
        sh_analysis_file = self.analysis_file[shared_columns]
        assert sh_local_ref.equals(sh_analysis_file)

    


"""Module to fetch Nextclade data."""

import subprocess
import os

def run_nextclade(command,  write=False):
    """Run Nextclade CLI.
    
    Parameters
    ----------
    command : list 
        Directory where test data will be downloaded.
    write : bool, default:False
        Determines whether to print output of command.
    
    Returns
    -------
    None
    
    """
 
    try:
        result = subprocess.run(command, check=True, capture_output=True, text=True)
        if write:
            print(result.stdout)
    except subprocess.CalledProcessError as e:
        print(f"Error running Nextclade: {e}")
    except FileNotFoundError:
        print("Nextclade executable not found. Make sure it is included in the system $PATH.")
    except Exception as e:
        print(f"An unexpected error occurred: {e}")

def list(full=False):
    """List available Nextclade datasets.

    Parameters
    ----------
    full : bool, default:False
        Determines if only names of datasets will be shown.
    
    Returns
    -------
    None

    """

    if full:
        list_command = "nextclade dataset list"
    else:
        list_command = "nextclade dataset list --only-names"
    run_nextclade(list_command.split(), write=True)

def get(id, dir='', version=''):
    """Download specified Nextclade dataset into target directory.
    
    Parameters
    ----------
    id : str 
        Name of the Nextclade dataset.
    dir : str
        Directory where dataset will be downloaded.
    
    Returns
    -------
    None
    
    """

    # Define target directory where data will be downloaded
    if isinstance(dir, str) and dir.strip(): # Use user-provided directory 
        target_dir = dir 
        if not os.path.isdir(target_dir):
            os.makedirs(target_dir)
    else: # Or use current working directory by default
        target_dir = os.getcwd()
    
    get_command = f"nextclade dataset get --name {id} --output-dir {target_dir}"
    if version != '':
        get_command = get_command + f" --tag {version}"
    run_nextclade(get_command.split())


"""Module that creates Nextclade command and captures Nextclade output."""

import subprocess
import os
import pandas as pd


def create_command(**kwargs):
    """Creates Nextclade command based on user input.
    
    Parameters
    ----------
    input : dict 
        Dictionary containing the user input to nextclade().
    secure_analysis_dir : str
        Secure temporary directory to store Nexclade analysis TSV output.
    secure_ref_dir : str
        Secure temporary directory to store reference sequence from Nextclade.

    Returns
    -------
    list
        A list whose elements constitute the Nextclade command.

    """

    # Getting user input
    input = kwargs["input"]
    seq = input["seq"]
    ref = input["ref"]   

    # Getting nextread-created directories
    secure_analysis_dir = kwargs["secure_analysis_dir"]
    secure_ref_dir = kwargs["secure_ref_dir"]

    # Creating command based on user input
    if os.path.isfile(ref): # Reference FASTA is provided
        nextclade_command = f"nextclade run -r {ref} -t {os.path.join(secure_analysis_dir, 'analysis.tsv')}".split()
        if "gene" in input: # Gene annotation is provided
            gene = input["gene"]
            nextclade_command = nextclade_command + ["-m", gene]
        nextclade_command = nextclade_command + [seq]
        gene_path = gene
    else: # Name of Nextclade reference is provided
        secure_ref_dir = kwargs["secure_analysis_dir"]
        # Get dataset first and then run the command
        gene_path = os.path.join(secure_ref_dir, 'genome_annotation.gff3')
        first_command = f"nextclade dataset get -n {ref} -o {secure_ref_dir}".split()
        second_command = f"nextclade run -r {os.path.join(secure_ref_dir, 'reference.fasta')} -m {gene_path} -t {os.path.join(secure_analysis_dir, 'analysis.tsv')} {seq}".split()
        nextclade_command = [first_command, second_command]
    return nextclade_command, gene_path


def capture_output(command):
    """Runs Nextclade CLI and captures the output.

    Parameters
    ----------
    command : str
        The Nextclade CLI command.

    Returns
    -------
    pandas.DataFrame
        A DataFrame of the Nextclade analysis TSV output.

    """
    try:
        if len(command) != 2: # Reference FASTA is provided
            subprocess.run(command, check=True)
            analysis_arg_ind = command.index("-t") + 1
            analysis_file_path = command[analysis_arg_ind]
        else: # Name of Nextclade reference is provided
            analysis_arg_ind = command[1].index("-t") + 1
            analysis_file_path = command[1][analysis_arg_ind]
            for line in command:
                subprocess.run(line, check=True)

        # Reading Nextclade analysis TSV output
        analysis_dataframe = pd.read_csv(analysis_file_path, delimiter='\t')
        # Removing index column
        analysis_dataframe.drop('index', axis=1, inplace=True)
        return analysis_dataframe
    except subprocess.CalledProcessError as e:
        print(f"Error running Nextclade: {e}")
    except FileNotFoundError:
        print("Nextclade executable not found. Make sure it is included in the system $PATH.")
    except Exception as e:
        print(f"An unexpected error occurred: {e}")


"""Tests whether processed data provided for plotting is correct."""

from create_profile_data import MyProfileData
from vargram import vargram
import matplotlib.pyplot as plt
import random
import pandas as pd
import tempfile
import os
import shutil
import pytest


@pytest.fixture(params=[(False, 0, 'counts'), (False, 10, 'counts'),
                        (False, 0, 'weights'), (False, 10, 'weights'),
                        (True, 0, 'counts'), (True, 10, 'counts'),
                        (True, 0, 'weights'), (True, 10, 'weights')])
def profile_data(request):
    """Feed input to VARGRAM based on different parameters and get output."""
    key_called, threshold, ytype = request.param
    num = random.randint(30,100)

    # Creating test outputs and inputs
    mbd = MyProfileData(key_called=key_called, num=num, ytype=ytype)
    output = mbd.create_output()
    input = mbd.create_input()

    # Getting vargram output
    vg = vargram(data=input, format='_test')
    vg.profile(threshold=threshold,ytype=ytype)
    if key_called == True:
        keys = mbd.create_keys()
        try:
            vargram_test_dir = tempfile.mkdtemp(prefix="vargram_test_dir")
            for i, key in enumerate(keys):
                key_path = os.path.join(vargram_test_dir, f'key_{i+1}.csv')
                key.to_csv(key_path, index=False)
                key.head(5)
                vg.key(key_path)
        finally:
            shutil.rmtree(vargram_test_dir)
    yield {"vg":vg, "output":output}
    plt.close()


class TestProfileData:

    def test_returned_data(self, profile_data):
        """Returned profile data should be equal to saved data. """ 
        vg = profile_data["vg"]
        result = vg.stat()
        try:
            vargram_test_dir = tempfile.mkdtemp(prefix="vargram_test_dir")
            saved_file_path = os.path.join(vargram_test_dir, 'saved_profile_data.csv')
            vg.save(saved_file_path, index=False)
            expected = pd.read_csv(saved_file_path)        
            assert result.equals(expected)
        finally:
            shutil.rmtree(vargram_test_dir)

    def test_batch_key_sums(self, profile_data):
        """The total sum of all batch and key columns should not be zero."""
        vg = profile_data["vg"]
        column_sums = vg.stat().sum(numeric_only=True, axis=1)
        zero_column_sums = column_sums.abs() < 1e-10
        result = zero_column_sums.any()
        expected = False
        assert result == expected
    
    def test_number_mutations(self, profile_data):
        """The length of the dataframe should be equal to the number of unique mutations.
        This is true for the test input created.
        """
        vg = profile_data["vg"]
        output = vg.stat()
        result = len(output['mutation'])
        expected = len(output['mutation'].unique())
        assert result == expected
    
    def test_order_mutations(self, profile_data):
        """Per gene, the positions should only increase as you go down the rows."""
        vg = profile_data["vg"]
        output = vg.stat()
        genes = output['gene'].unique().tolist()

        orders = []
        for gene in genes:
            gene_output = output[output['gene'] == gene]
            positions = list(gene_output['position'])
            position_increasing = all(positions[i] < positions[i + 1] for i in range(len(positions) - 1))
            orders.append(position_increasing)
        
        result = all(orders)
        expected = True
        assert result == expected

    def test_whole_data(self, profile_data):
        """The returned profile data should be equal to expected profile data."""
        vg = profile_data["vg"]
        expected = profile_data["output"]
        result = vg.stat()
        assert result.equals(expected) 


"""Main module for generating VARGRAM figures and data."""

from .wranglers._wrangler import Wrangler
from .plots._profile import Profile
import pandas as pd
import os


class vargram:

    def __init__(self, **vargram_kwargs):
        """Initializes the VARGRAM object, getting data 
        and joining it with metadata (if provided).

        Parameters
        ----------
        seq : str
            FASTA file path of the sequences.
        ref : str
            FASTA file path of the reference sequence.
        gene : str
            GFF3 file path of the genome annotation.
        data : str or pandas.DataFrame
            The data to be plotted.
        metadata : string or pandas.DataFrame
            The metadata to be joined with the data.
        join : str or list
            Provides the column to do an outer merge on. 
            If data and metadata do not share a column, 
            the two column names which share the values of interest may be provided.
        
        Returns
        -------
        None

        Raises
        ------
        ValueError
            If neither 'seq' nor 'data' is provided.
            If 'metadata' is provided but 'join' is not.
            If the resulting '_data' attribute is empty.

        """
        # Defining initial values
        self._initialize_variables()
        self._vargram_kwargs = vargram_kwargs
    
    def _initialize_variables(self):
        """Sets initial values of attributes."""
        self._methods_called = [] # List of all methods called
        self._methods_kwargs = [] # List of the arguments of the corresponding methods called
        self._plots = ['profile'] # These are the 'plot methods'
        self._terminals = ['show', 'save', 'stat'] # These are the 'terminal methods'
        self._generate_plot = False
        self._plots_indices = []
        self._clean_keys()
        self._shown = False
        self._saved = False
    
    def _clean_keys(self):
        """Flushes key variables clean."""
        self._master_key_data = pd.DataFrame()
        self._nkeys = 0
        self._key_labels = []
        self._key_colors = []
    
    def _generate(self):
        """Runs all called methods in correct order"""

        # Ensure aes method is called
        if '_aes' not in self._methods_called[self._latest_plot_index:]:
            self.aes() # Call aes to get default arguments
            # Move the aes method name and arguments next to the plot method called
            self._methods_called.insert(self._latest_plot_index + 1, '_aes')
            self._methods_kwargs.insert(self._latest_plot_index + 1, self._methods_kwargs[-1])
            self._methods_called.pop()
            self._methods_kwargs.pop()
        
        # Getting group of methods and corresponding arguments called since last called plot
        latest_method_calls = self._methods_called[self._latest_plot_index:]
        latest_method_kwargs = self._methods_kwargs[self._latest_plot_index:]

        # Determining whether to generate plots or not
        if not self._generate_plot:
            getattr(self, latest_method_calls[-1])(**self._methods_kwargs[-1])
            return
        
        # Creating plot object instance
        plot_class = latest_method_calls[0][1:].title() 
        plot_object = globals()[plot_class]
        self._vargram_kwargs['plot'] = plot_class
        wrangled_data = Wrangler(self._vargram_kwargs).get_wrangled_data()
        self._plot_instance = plot_object(wrangled_data)

        # Rearranging so that auxiliary methods are run before plot and save/show methods
        latest_method_calls.append(latest_method_calls[-1])
        latest_method_calls[-2] = latest_method_calls[0]
        latest_method_calls.pop(0)
        latest_method_kwargs.append(latest_method_kwargs[-1])
        latest_method_kwargs[-2] = latest_method_kwargs[0]
        latest_method_kwargs.pop(0)

        # Generating figure based on most recent plot called
        run_key = True
        for i, method in enumerate(latest_method_calls):
            # Running each method since the most recent plot called
            if run_key or method != '_key': # Ensuring that the internal key method is only called once
                if method == '_key':
                    run_key = False
                if method == '_profile':
                    self._plot_data = getattr(self, method)(**latest_method_kwargs[i])
                else:
                    getattr(self, method)(**latest_method_kwargs[i])
            else:
                continue
        self._generate_plot = False
    
    def _show(self, **_show_kwargs): 
        """Show generated figure"""
        getattr(self._plot_instance, 'show')()
    
    def _save(self, **_save_kwargs):
        """Save generated figure or data"""
        getattr(self._plot_instance, 'save')(**_save_kwargs)
    
    def _stat(self, **_stat_kwargs):
        """Kept for consistency with _generate() flow."""
        pass
    
    def _profile(self, **_profile_kwargs):
        """Process data for plotting"""
        return getattr(self._plot_instance, 'process')(**_profile_kwargs)

    def _key(self, **_key_kwargs):
        """Process key data for plotting"""
        getattr(self._plot_instance, 'key')(key_data=self._master_key_data, 
                                            key_labels=self._key_labels, 
                                            key_colors=self._key_colors)
    
    def _struct(self, **_struct_kwargs):
        """Modify aesthetic attributes"""
        getattr(self._plot_instance, 'struct_method')(**_struct_kwargs)

    def _aes(self, **_aes_kwargs):
        """Modify aesthetic attributes"""
        getattr(self._plot_instance, 'aes')(**_aes_kwargs)
    
    def stat(self):
        """Wrapper for stat method
        
        Returns
        -------
        None

        """
        self._methods_called.append('_stat')
        self._methods_kwargs.append({'empty_string':''}) 
        # The unused empty string argument is so as to be able to maintain 
        # length of methods and methods_kwargs the same
        self._generate()
        return self._plot_data
        
    def show(self): 
        """Wrapper for show method"""
        self._shown = True
        self._methods_called.append('_show')
        self._methods_kwargs.append({'empty_string':''}) 
        # The unused empty string argument is so as to be able to maintain
        # length of methods and methods_kwargs the same
        self._generate()

    def save(self, fname, **save_kwargs):
        """Wrapper for save method"""
        self._saved = True
        save_kwargs['fname'] = fname
        self._methods_called.append('_save')
        self._methods_kwargs.append(save_kwargs)
        self._generate()

    def profile(self,
                threshold=10, 
                x='mutation', 
                y='',
                ytype='', 
                group='gene', 
                stack='batch',
                ):
        """Captures profile method arguments
        
        Parameters
        ----------
        x : str, default:'mutation'
            The x-axis variable (column name of data).
        y : str
            The y-axis variable (column name of data).
        ytype : {'counts', 'weights'}
            The type of calculation for the y-axis.
            'counts' indicate raw counts.
            'weights' indicate normalization by total count.
        group : str, default:'gene'
            The column name for the groups.
        stack : str, default:'batch'
            The column name for the stacks.
        threshold : int, default:50
            The minimum number of occurences of an x value to be included.

        Returns
        -------
        None

        """
        profile_kwargs = locals()
        del profile_kwargs['self']
        self._methods_called.append('_profile')
        self._methods_kwargs.append(profile_kwargs)
        self._latest_plot_index = len(self._methods_called) - 1
        self._generate_plot = True
        self._clean_keys()

    def key(self, key_data, x='mutation', group='gene', label='', color = '#5E5E5E'):
        """Joins all key data.
        
        Parameters
        ----------
        key_data : str
            File path of key lineage
        x : str, default:'mutation'
            The x-axis variable (column name of data).
        group : str, default:'gene'
            The column name for the groups.
        label : str
            The name of the key lineage.
        color : str
            The color of key mutations on the generated heatmap.
        
        Returns
        -------
        None

        """
        self._methods_called.append('_key')
        self._methods_kwargs.append({'empty_string':''}) 
        # The unused empty string argument is so as to be able to maintain 
        # length of methods and methods_kwargs the same

        self._nkeys += 1
        # Reading data
        if isinstance(key_data, str):
            key_read = self._read_comma_or_tab(key_data)
        if isinstance(key_data, pd.DataFrame):
            key_read = key_data.copy()

        # Just keeping the 'x' and 'group' columns
        key_df = key_read[[group, x]].copy()

        # Getting name of key
        if label == '' and isinstance(key_data, str):
            label = os.path.basename(key_data)
            label = os.path.splitext(label)[0]
        self._key_labels.append(label)

        # Getting color of key
        self._key_colors.append(color)

        # Gathering in one master dataframe
        if self._nkeys > 1:
            key_df[label] = 1
            self._master_key_data = pd.merge(self._master_key_data, key_df, 
                                             on=[group, x], how='outer')
            self._master_key_data.fillna(0, inplace=True)
            self._master_key_data.reset_index(drop=True, inplace=True)
        else:
            self._master_key_data = pd.DataFrame()
            self._master_key_data[group] = key_df[group]
            self._master_key_data[x] = key_df[x]
            self._master_key_data[label] = 1

    def struct(self, struct_arg):
        """Retains only groups that are provided.
        
        Parameters
        ----------
        struct_arg : str
            A formatted string indicating the groups to retain per row.
            Commas separate group names and forward slashes separate rows.

        Returns
        -------
        None

        """
        self._methods_called.append('_struct')
        self._methods_kwargs.append({'struct_key':struct_arg})

    def aes(self,
            order=False, 
            flat=False,
            stack_label=[], 
            stack_title='', 
            stack_color='', 
            group_title='', 
            legtitle_fontsize='large', 
            legentry_fontsize='large',
            xticks_fontsize=6, 
            xticks_rotation=90,
            yticks_fontsize=10,
            ylabel='', 
            ylabel_fontsize='large', 
            group_fontsize='large',
            key_fontsize='medium',
            figsize=None,
            aspect=None
            ):
        """Captures modification of aesthetic attributes.
        
        Parameters
        ----------
        order : bool, default:False
            Determines whether genes will be ordered according to start position.
        flat : bool, default:False
            Determines whether genes will be shown on a horizontal layout.
        stack_label : str
            Name(s) of the stack(s).
        stack_title : str
            Legend title for the stack(s).
        stack_color : str or list
            Color(s) of the stack(s).
        group_title : str
            Legend title for the groups.
        legtitle_fontsize : str or float
            The fontsize of the legend titles.
        legentry_fontsize : str or float
            The fontsize of the legend entries.
        xticks_fontsize : float
            The fontsize of the x-axis ticks.
        xticks_rotation : float
            The degree of rotation of the x-axis ticks.
        yticks_fontsize : float
            The fontsize of the y-axis ticks.
        ylabel : str
            The y-axis label.
        ylabel_fontsize : str or float
            The fontsize of the y-axis label.
        group_fontsize : str or float
            The fontsize of the group label.
        key_fontsize : str or float
            The fontsize of the heatmap (key) row labels.
        figsize : tuple
            Tuple containing the width and height of the figure in inches.
        aspect : float
            The aspect ratio of the figure.

        Returns
        -------
        None

        """
        aes_kwargs = locals()
        del aes_kwargs['self']
        self._methods_called.append('_aes')
        self._methods_kwargs.append(aes_kwargs)

    def _read_comma_or_tab(self, file):
        """Reads a CSV or TSV file.

        Parameters
        ----------
        file : str
            The file path to be read.
        
        Returns
        -------
        pandas.DataFrame
            The DataFrame of the file.
        
        Raises
        ------
        ValueError
            If the given file is not a CSV or TSV file.

        """
        root, extension = os.path.splitext(file)
        if extension == '.csv':
            return pd.read_csv(file)
        elif extension == '.tsv':
            return pd.read_csv(file, delimiter='\t')
        else:
            raise ValueError("Incorrect file format. Expecting CSV or TSV file.")        